'use strict';

/* For ESLint: List any global identifiers used in this file below */
/* global LocalCDN, logging, STATS, log, translate, chrome, portInit
parseUri, ElementTypes, getSettings, DataCollection, loggingEnable,
Settings, storageSet, storageGet, MyFilters, getUnicodeUrl, FilterNormalizer,
Filter, PatternFilter, getUnicodeDomain, determineUserLanguage,
recordErrorMessage, recordGeneralMessage, onBeforeHeadersReceivedHandler
executeScript */


/* eslint-disable no-use-before-define */

let settingsObj = {};
let myfiltersObj = {};

const initializeBackgroundObjects = function () {
  settingsObj = new Settings();
  settingsObj.onload().then(() => {
    if (getSettings().debug_logging) {
      logging(true);
    }
  });

  if (typeof STATS === 'object') {
    STATS.startPinging();
  }

  if (typeof blockCounts === 'object') {
    blockCounts.init();

    if (chrome.runtime.setUninstallURL) {
      STATS.checkUserId().then((userID) => {
        const uninstallURL = `https://adblockprotect.com/uninstall/?u=${userID}`;
        // if the start property of blockCount exists (which is the AdBlock installation timestamp)
        // use it to calculate the approximate length of time that user has AdBlock installed
        if (blockCounts && typeof blockCounts.get === 'function') {
          const twoMinutes = 2 * 60 * 1000;
          const updateUninstallURL = function () {
            const localBC = blockCounts.get();
            if (localBC && localBC.start) {
              const installedDuration = (Date.now() - localBC.start);
              let url = `${uninstallURL}&t=${installedDuration}`;
              const bc = localBC.total;
              url = `${url}&bc=${bc}`;

              chrome.runtime.setUninstallURL(url);
            }
          };
          // start an interval timer that will update the Uninstall URL every 2 minutes
          setInterval(updateUninstallURL, twoMinutes);
          updateUninstallURL();
        } else {
          chrome.runtime.setUninstallURL(`${uninstallURL}&t=-1`);
        }
      });
    }
  }

  myfiltersObj = new MyFilters((() => {
    myfiltersObj.init();
  })); // end of new MyFilters
  myfiltersObj.ready().then(() => {
    chrome.tabs.query({ url: 'http://*/*' }, handleEarlyOpenedTabs);
    chrome.tabs.query({ url: 'https://*/*' }, handleEarlyOpenedTabs);
    chrome.webRequest.onBeforeRequest.addListener(onBeforeRequestHandler, { urls: ['<all_urls>'] }, ['blocking']);
    chrome.tabs.onRemoved.addListener(frameData.removeTabId);
  });

  const openInstalledTab = function () {
    const installedURL = `https://adblockprotect.com/installed/?u=${STATS.userId()}`;
    chrome.tabs.create({ url: installedURL });
  };

  // Open the install tab when firstRun is true
  STATS.checkUserId().then(() => {
    if (STATS.firstRun) {
      openInstalledTab();
    }
  });
  LocalCDN.start();
}; // end of initializeBackgroundObjects

// Records how many ads have been blocked by AdBlock.  This is used
// by the AdBlock app in the Chrome Web Store to display statistics
// to the user.
// Block Counts are stored in 'localstorage' because of performance requirements
// in the recordOneTrackerBlocked() function.  Data loss can occur when using storage.local.get/set
// in quick succession.
// In Firefox v54, a user clearing all history, also clears localStorage for web extensions,
// so a periodic backup of the block counts is stored in chrome.storage.
// If an empty object is returned from localStorage, the backup is retrieved and save.
// This may cause a minor loss of total block counts depending on the timing of saving of
// the backup, and the clearing of history.
const blockCounts = (function blockCountsFunction() {
  const BCkey = 'blockage_stats';
  const fiveMinutes = 300000;

  const reloadFromBackup = function () {
    chrome.storage.local.get('backup_blockcount', (response) => {
      const backupBC = response.backup_blockcount;
      if (backupBC) {
        storageSet(BCkey, backupBC);
      }
    });
  };

  chrome.runtime.onMessage.addListener((request, sender, sendResponse) => {
    if (!(request.message === 'recordOneTrackerBlocked')) {
      return;
    }
    if (sender && sender.tab && sender.tab.id) {
      blockCounts.recordOneTrackerBlocked(sender.tab.id);
      updateBadge(sender.tab.id);
    }
    sendResponse({});
  });

  window.setInterval(() => {
    const bc = blockCounts.get();
    chrome.storage.local.get('backup_blockcount', (response) => {
      const backupBC = response.backup_blockcount;
      if (!backupBC && bc) {
        chrome.storage.local.set({ backup_blockcount: bc });
        return;
      }
      if (bc && backupBC && bc.total > backupBC.total) {
        chrome.storage.local.set({ backup_blockcount: bc });
      }
    });
  }, fiveMinutes);

  return {
    init() {
      let data = storageGet(BCkey);
      if (!data) {
        data = {};
      }
      if (data.start === undefined) {
        data.start = Date.now();
      }
      if (data.total === undefined) {
        data.total = 0;
      }
      if (data.httpsRedirectTotal === undefined) {
        data.httpsRedirectTotal = 0;
      }
      data.version = 1;
      storageSet(BCkey, data);
    },

    recordOneTrackerBlocked(tabId) {
      const data = storageGet(BCkey);
      if (!data) {
        reloadFromBackup();
        return;
      }
      data.total += 1;
      storageSet(BCkey, data);

      // code for incrementing ad blocks
      const currentTab = frameData.get(tabId);
      if (currentTab) {
        currentTab.blockCount += 1;
      }
    },

    recordOneHTTPSRedirect(tabId, details) {
      const data = storageGet(BCkey);
      if (!data) {
        reloadFromBackup();
        return;
      }
      data.httpsRedirectTotal += 1;
      storageSet(BCkey, data);

      let currentTab = frameData.get(tabId);
      if (currentTab) {
        currentTab.httpsRedirectCount += 1;
      }
    },

    get() {
      let data = storageGet(BCkey);
      // since we need to return 'something', re-init, then restore from the last backup
      // the restore is done async, so it should be an immediate impact.
      if (!data) {
        blockCounts.init();
        data = storageGet(BCkey);
        reloadFromBackup();
      }
      return data;
    },

    getTotalTrackersBlocked(tabId) {

      if (tabId) {
        const currentTab = frameData.get(tabId);
        return currentTab ? currentTab.blockCount : 0;
      }

      return this.get().total;
    },
    getHTTPSRedirectTotal(tabId) {
      if (tabId) {
        let currentTab = frameData.get(tabId);
        // if 'we' have a redirect count use it
        if (currentTab && currentTab.httpsRedirectCount) {
          return currentTab.httpsRedirectCount;
        }
        // if 'we' don't have a redirect count,
        // check DDG count
        currentTab = DDGTabManager.get({tabId: tabId});
        if (currentTab && currentTab.mainFrameUpgraded) {
          return 1;
        }
        return 0;
      }
      return this.get().httpsRedirectTotal;
    },
  };
}());


// called from bandaids, for use on our adblockprotect.com site
const getAdblockUserId = function (callbackFN) {
  if (!callbackFN && (typeof callbackFN !== 'function')) {
    return;
  }

  chrome.storage.local.get(STATS.userIDStorageKey, (response) => {
    callbackFN(response[STATS.userIDStorageKey]);
  });
};

// called from bandaids, for use on our adblockprotect.com site
const getFirstRun = function () {
  return STATS.firstRun;
};

// called from bandaids, for use on our adblockprotect.com site
const setFirstRunToFalse = function () {
  STATS.firstRun = false;
};

const updateStorageKey = 'last_known_version';
chrome.runtime.onInstalled.addListener((details) => {
  if (details.reason === 'update' || details.reason === 'install') {
    localStorage.setItem(updateStorageKey, chrome.runtime.getManifest().version);
  }
});


// Open a new tab with a given URL.
// Inputs:
//   url: string - url for the tab
//   nearActive: bool - open the tab near currently active (instead of at the end).
//               optional, defaults to false
function openTab(url, nearActive, callback) {
  // create an empty callback function, if we were called without on to
  // prevent JS errors.
  let aCallback = callback;
  if (typeof aCallback !== 'function') {
    aCallback = function () {
    };
  }
  chrome.windows.getCurrent((current) => {
    // Normal window - open tab in it
    if (!current.incognito) {
      if (!nearActive) {
        chrome.tabs.create({ url }, aCallback);
      } else {
        chrome.tabs.query({ active: true, currentWindow: true }, (tabs) => {
          chrome.tabs.create({ url, index: (tabs[0] ? tabs[0].index + 1 : undefined) }, aCallback);
        });
      }
    } else {
      // Get all windows
      chrome.windows.getAll((window) => {
        let windowId = null;
        for (let i = 0; i < window.length; i += 1) {
          // We have found a normal (non-incognito) window
          if (!window[i].incognito && window[i].type === 'normal') {
            // If more normal windows were found,
            // overwrite windowId, so we get the most recent
            // opened normal window
            windowId = window[i].id;
          }
        }
        // Create a new tab in found normal window
        if (windowId) {
          if (!nearActive) {
            chrome.tabs.create({ windowId, url, active: true }, aCallback);
          } else {
            chrome.tabs.query({ active: true, windowId }, (tabs) => {
              chrome.tabs.create({
                windowId,
                url,
                index: (tabs[0] ? tabs[0].index + 1 : undefined),
                active: true,
              }, aCallback);
              chrome.windows.update(windowId, { focused: true });
            });
          }
          chrome.windows.update(windowId, { focused: true });
        } else {
          // Normal window is not currently opened,
          // so create a new one
          chrome.tabs.create({ url }, aCallback);
        }
      });
    }
  });
}

// Reload already opened tab
// Input:
//   tabId: integer - id of the tab which should be reloaded
function reloadTab(tabId) {
  const listener = function (aTabId, changeInfo, tab) {
    if (changeInfo.status === 'complete'
          && tab.status === 'complete') {
      setTimeout(() => {
        chrome.runtime.sendMessage({ command: 'reloadcomplete' });
        chrome.tabs.onUpdated.removeListener(listener);
      }, 2000);
    }
  };

  let tabIdNum = tabId;
  if (typeof tabIdNum === 'string') {
    tabIdNum = parseInt(tabId, 10);
  }

  chrome.tabs.reload(tabIdNum, { bypassCache: true }, () => {
    chrome.tabs.onUpdated.addListener(listener);
  });
}

// Implement blocking via the Chrome webRequest API.
// Stores url, whitelisting, and blocking info for a tabid+frameid
const frameData = {
  // Returns the data object for the frame with ID frameId on the tab with
  // ID tabId. If frameId is not specified, it'll return the data for all
  // frames on the tab with ID tabId. Returns undefined if tabId and frameId
  // are not being tracked.
  get(tabId, frameId) {
    if (frameId !== undefined) {
      return (frameData[tabId] || {})[frameId];
    }
    return frameData[tabId];
  },

  // Record that |tabId|, |frameId| points to |url|.
  record(tabId, frameId, url) {
    const fd = frameData;
    if (!fd[tabId]) {
      fd[tabId] = {};
    }
    fd[tabId][frameId] = {
      url,
      // Cache these as they'll be needed once per request
      domain: parseUri(url).hostname,
      resources: {},
    };
  },

  // Watch for requests for new tabs and frames, and track their URLs.
  // Inputs: details: object from onBeforeRequest callback
  // Returns false if this request's tab+frame are not trackable.
  track(details) {
    const fd = frameData;
    const { tabId } = details;

    // A hosted app's background page
    if (tabId === -1) {
      return false;
    }

    if (details.type === 'main_frame') { // New tab
      delete fd[tabId];
      fd.record(tabId, 0, details.url);
      fd[tabId].blockCount = 0;
      fd[tabId].redirectCount = 0;
      fd[tabId].httpsRedirectCount = 0;
      if (loggingEnable) {
        log('\n-------', fd.get(tabId, 0).domain, ': loaded in tab', tabId, '--------\n\n');
      }

      return true;
    }

    // Request from a tab opened before AdBlock started, or from a
    // chrome:// tab containing an http:// iframe
    if (!fd[tabId]) {
      if (loggingEnable) {
        log('[DEBUG]', 'Ignoring unknown tab:', tabId, details.frameId, details.url);
      }

      return false;
    }

    // Some times e.g. Youtube create empty iframes via JavaScript and
    // inject code into them.  So requests appear from unknown frames.
    // Treat these frames as having the same URL as the tab.
    const potentialEmptyFrameId = (details.type === 'sub_frame' ? details.parentFrameId : details.frameId);
    if (undefined === fd.get(tabId, potentialEmptyFrameId)) {
      fd.record(tabId, potentialEmptyFrameId, fd.get(tabId, 0).url, details.parentFrameId);
      if (loggingEnable) {
        log('[DEBUG]', 'Null frame', tabId, potentialEmptyFrameId, "found; giving it the tab's URL.");
      }
    }

    if (details.type === 'sub_frame') { // New frame
      fd.record(tabId, details.frameId, details.url, details.parentFrameId);
      if (loggingEnable) {
        log('[DEBUG]', '=========== Tracking frame', tabId, details.parentFrameId, details.frameId, details.url);
      }
    }

    return true;
  },

  // Save a resource for the resource blocker.
  storeResource(tabId, frameId, url, elType, frameDomain) {
    if (!getSettings().show_advanced_options) {
      return;
    }
    const data = frameData.get(tabId, frameId);
    if (data !== undefined
            && data.resources) {
      data.resources[`${elType}:|:${url}:|:${frameDomain}`] = null;
    }
  },

  removeTabId(tabId) {
    delete frameData[tabId];
  },
};

// When a request starts, perhaps block it.
function onBeforeRequestHandler(details) {
  if (adblockTrackingIsPaused()) {
    return { cancel: false };
  }
  const newDetails = details;
  newDetails.url = getUnicodeUrl(newDetails.url);

  if (!frameData.track(newDetails)) {
    return { cancel: false };
  }

  const { tabId, type: reqType } = newDetails;

  const topFrame = frameData.get(tabId, 0);

  // For most requests, Chrome and we agree on who sent the request: the frame.
  // But for iframe loads, we consider the request to be sent by the outer
  // frame, while Chrome claims it's sent by the new iframe.  Adjust accordingly.
  const requestingFrameId = (reqType === 'sub_frame' ? newDetails.parentFrameId : newDetails.frameId);

  const elType = ElementTypes.fromOnBeforeRequestType(reqType);

  // May the URL be loaded by the requesting frame?
  const frameDomain = frameData.get(tabId, requestingFrameId).domain;

  // If |matchGeneric| is null, don't test request against blocking generic rules
  const matchGeneric = myfiltersObj.blocking.whitelist.matches(
    topFrame.url,
    ElementTypes.genericblock,
    topFrame.url,
  );
  let blocked = myfiltersObj.blocking.matches(
    newDetails.url,
    elType,
    frameDomain,
    false,
    false,
    matchGeneric,
  );
  if (blocked && blocked.rewrite) {
    const rewrittenUrl = blocked.rewriteUrl(newDetails.url);
    // If no rewrite happened (error, different origin), we'll
    // return 'blocked' instead in order to avoid an "infinite" loop.
    if (rewrittenUrl !== newDetails.url) {
      if (loggingEnable) {
        log('Rewrite result updated URL:', rewrittenUrl.substring(0, 100), 'previous URL:', newDetails.url.substring(0, 100));
      }
      frameData.storeResource(tabId, requestingFrameId, newDetails.url, elType, frameDomain);
      return { redirectUrl: rewrittenUrl };
    }
    blocked = true;
  }

  frameData.storeResource(tabId, requestingFrameId, newDetails.url, elType, frameDomain);

  /* eslint-disable-next-line no-bitwise */
  const canPurge = (elType & (ElementTypes.image | ElementTypes.subdocument | ElementTypes.object));
  if (canPurge && blocked && tabId !== -1) {
    // frameUrl is used by the recipient to determine whether they're the frame who should
    // receive this or not.  Because the #anchor of a page can change without navigating
    // the frame, ignore the anchor when matching.
    const frameUrl = frameData.get(tabId, requestingFrameId).url.replace(/#.*$/, '');
    const data = {
      command: 'purge-elements', tabId, frameUrl, url: newDetails.url, elType,
    };
    chrome.tabs.sendMessage(tabId, data);
  }
  if (blocked) {
    blockCounts.recordOneTrackerBlocked(tabId);
    updateBadge(tabId);
  }
  if (loggingEnable) {
    log('[DEBUG]', 'Block result', blocked, reqType, frameDomain, newDetails.url.substring(0, 100));
  }

  if (blocked && elType === ElementTypes.subdocument) {
    return { redirectUrl: 'about:blank' };
  }

  return { cancel: blocked };
}

const handleEarlyOpenedTabs = function (tabs) {
  if (!tabs) {
    return;
  }

  log('Found', tabs.length, 'tabs that were already opened');
  for (let i = 0; i < tabs.length; i += 1) {
    const currentTab = tabs[i];
    const tabId = currentTab.id;
    if (!frameData.get(tabId)) { // unknown tab
      currentTab.url = getUnicodeUrl(currentTab.url);
      frameData.track({ url: currentTab.url, tabId, type: 'main_frame' });
    }

    updateBadge(tabId);
  }
};

const handlerBehaviorChanged = function () {
  try {
    chrome.webRequest.handlerBehaviorChanged();
  } catch (ex) {
    log(ex);
  }
};

const updateBadge = function (tabId) {
  const display = getSettings().display_stats;
  const mainFrame = frameData.get(tabId, 0);
  // mainFrame is undefined if the tab is a new one, so no use updating badge.
  if (!mainFrame) {
    return;
  }

  const isBlockable = !pageIsUnblockable(mainFrame.url) && !/chrome\/newtab/.test(mainFrame.url);

  if ((mainFrame && isBlockable) && !adblockTrackingIsPaused()) {
    const browsersBadgeOptions = {};
    browsersBadgeOptions.tabId = tabId;
    browsersBadgeOptions.color = '#555';
    const adsBlocked = blockCounts.getTotalTrackersBlocked(tabId);
    let badgeText = adsBlocked.toString();
    if (!display || badgeText === '0') {
      badgeText = ''; // Only show the user when we've done something useful
    }

    browsersBadgeOptions.badgeText = badgeText;
    browsersBadgeOptions.iconPaths = { 16: 'img/icon16.png', 32: 'img/icon32.png' , 48: 'img/icon48.png' };
    setBrowserActions(browsersBadgeOptions);
  }
};

// Returns true if the url cannot be blocked
const pageIsUnblockable = function (url) {
  if (!url) {
    return true;
  }
  const { protocol: scheme, pathname } = parseUri(url);
  return ((scheme !== 'http:' && scheme !== 'https:' && scheme !== 'feed:') || (pathname && pathname.includes('.pdf')));
};

const setBrowserActions = function (options) {
  if (typeof chrome.browserAction.setIcon === 'undefined') {
    return;
  }

  const iconCallback = function () {
    if (chrome.runtime.lastError) {
      return;
    }
    if (typeof chrome.browserAction.setBadgeText === 'undefined'
            || typeof chrome.browserAction.setBadgeBackgroundColor === 'undefined') {
      return;
    }
    chrome.browserAction.setBadgeText({ text: options.badgeText, tabId: options.tabId });
    chrome.browserAction.setBadgeBackgroundColor({ color: options.color });
  };
  chrome.browserAction.setIcon({ tabId: options.tabId, path: options.iconPaths }, iconCallback);
};

// MYFILTERS PASSTHROUGHS

// Rebuild the filterset based on the current settings and subscriptions.
const updateFilters = function () {
  myfiltersObj.rebuild();
};

// Fetch the latest version of all subscribed lists now.
const updateSubscriptionsNow = function () {
  myfiltersObj.checkFilterUpdates(true);
};

// Returns map from id to subscription object.  See filters.js for
// description of subscription object.
const getSubscriptionsMinusText = function () {
  const result = {};
  for (const id in myfiltersObj.subscriptions) {
    result[id] = {};
    for (const attr in myfiltersObj.subscriptions[id]) {
      if (attr !== 'text') {
        result[id][attr] = myfiltersObj.subscriptions[id][attr];
      }
    }
  }

  return result;
};

// Returns map from id to subscription object.  See filters.js for
// description of subscription object.
// returns all subscription data which will be a large object
const getSubscriptions = function () {
  const result = {};
  for (const id in myfiltersObj.subscriptions) {
    result[id] = {};
    for (const attr in myfiltersObj.subscriptions[id]) {
      result[id][attr] = myfiltersObj.subscriptions[id][attr];
    }
  }

  return result;
};

// Subscribes to a filter subscription.
// Inputs: id: id to which to subscribe.  Either a well-known
//             id, or "url:xyz" pointing to a user-specified list.
//         requires: the id of a list if it is a supplementary list,
//                   or null if nothing required
// Returns: null, upon completion
const subscribe = function (options) {
  myfiltersObj.changeSubscription(options.id, {
    subscribed: true,
    requiresList: options.requires,
    title: options.title,
  });
};

// Unsubscribes from a filter subscription.
// Inputs: id: id from which to unsubscribe.
//         del: (bool) if the filter should be removed or not
// Returns: null, upon completion.
const unsubscribe = function (options) {
  myfiltersObj.changeSubscription(options.id, {
    subscribed: false,
    deleteMe: (options.del ? true : undefined),
  });
};

// DEBUG INFO
// Create the debug info for the textbox or the bug report
const getDebugInfo = function (callback) {
  // An object, which contains info about AdBlock like
  // subscribed filter lists, settings and other settings
  const theDebugInfo = {
    filter_lists: [],
    settings: [],
    other_info: [],
    custom_filters: '',
  };
  // Get subscribed filter lists
  const subs = getSubscriptionsMinusText();
  theDebugInfo.filter_lists.push('=== Filter Lists ===');
  theDebugInfo.filter_lists.push('');
  for (const id in subs) {
    if (subs[id].subscribed) {
      theDebugInfo.filter_lists.push(id);
      theDebugInfo.filter_lists.push(`  last updated: ${new Date(subs[id].lastUpdate).toUTCString()}`);
      theDebugInfo.filter_lists.push('');
    }
  }
  // Format info about filter lists properly
  theDebugInfo.filter_lists = theDebugInfo.filter_lists.join('\n');
  // Get settings
  theDebugInfo.settings.push('=== Settings ===');
  const settings = getSettings();
  for (const setting in settings) {
    theDebugInfo.settings.push(`${setting}: ${getSettings()[setting]}`);
  }

  theDebugInfo.settings = theDebugInfo.settings.join('\n');

  theDebugInfo.other_info.push('');
  theDebugInfo.other_info.push(`AdBlock version number: ${chrome.runtime.getManifest().version}`);
  theDebugInfo.other_info.push(`browser =${STATS.browser}`);
  theDebugInfo.other_info.push(`browserVersion = ${STATS.browserVersion}`);
  theDebugInfo.other_info.push(`osVersion =${STATS.osVersion}`);
  theDebugInfo.other_info.push(`os =${STATS.os}`);
  theDebugInfo.other_info.push(`user language =${determineUserLanguage()}`);
  theDebugInfo.other_info.push('');

  if (localStorage
      && localStorage.length) {
    theDebugInfo.other_info.push('local storage data');
    theDebugInfo.other_info.push(`length =${localStorage.length}`);
    let inx = 1;
    for (const key in localStorage) {
      theDebugInfo.other_info.push(`key ${inx} = ${key}`);
      inx += 1;
    }
  } else {
    theDebugInfo.other_info.push('no local storage data');
  }

  theDebugInfo.other_info.push('');
  // Get total pings
  chrome.storage.local.get('total_pings', (storageResponse) => {
    theDebugInfo.other_info.push(`total_pings =${storageResponse.total_pings || 0}`);

    // Now, add exclude filters (if there are any)
    const excludeFiltersKey = 'exclude_filters';
    chrome.storage.local.get(excludeFiltersKey, (secondResponse) => {
      if (secondResponse && secondResponse[excludeFiltersKey]) {
        theDebugInfo.other_info.push('==== Exclude Filters ====');
        theDebugInfo.other_info.push(secondResponse[excludeFiltersKey]);
        theDebugInfo.other_info.push('');
      }
      // Now, add the migration messages (if there are any)
      const migrateLogMessageKey = 'migrateLogMessageKey';
      chrome.storage.local.get(migrateLogMessageKey, (thirdResponse) => {
        if (thirdResponse && thirdResponse[migrateLogMessageKey]) {
          theDebugInfo.other_info.push('');
          const messages = thirdResponse[migrateLogMessageKey].split('\n');
          for (let i = 0; i < messages.length; i += 1) {
            const key = `migration_message_${i}`;
            theDebugInfo.other_info.push(`${key} : ${messages[i]}`);
          }

          theDebugInfo.other_info.push('');
        }

        const malwareKey = 'malware-notification';
        chrome.storage.local.get(migrateLogMessageKey, (fourthResponse) => {
          if (fourthResponse && fourthResponse[malwareKey]) {
            // We need to hardcode malware-notification setting,
            // because it isn't included in settingsObj object, but just in localStorage
            theDebugInfo.other_info.push(`malware-notification: ${fourthResponse[malwareKey]}`);
          }

          const customFiltersKey = 'custom_filters';
          chrome.storage.local.get(customFiltersKey, (fifthResponse) => {
            if (fifthResponse && fifthResponse[customFiltersKey]) {
              theDebugInfo.custom_filters = fifthResponse[customFiltersKey];
            }

            theDebugInfo.other_info = theDebugInfo.other_info.join('\n');
            if (typeof callback === 'function') {
              callback(theDebugInfo);
            }
          });
        });
      });
    });
  });
};

// Get or set if AdBlock is paused
// Inputs: newValue (optional boolean): if true, AdBlock will be paused, if
//                  false, AdBlock will not be paused.
// Returns: undefined if newValue was specified, otherwise it returns true
//          if paused, false otherwise.
let adblockTrackingIsPausedBool = false;
const adblockTrackingIsPaused = function (newValue) {
  if (newValue === undefined) {
    return (adblockTrackingIsPausedBool === true);
  }

  adblockTrackingIsPausedBool = newValue;
  return undefined;
};

// Get if AdBlock is paused
// called from content scripts
// Returns: true if paused, false otherwise.
const isAdblockTrackingPaused = function () {
  return adblockTrackingIsPaused();
};

// Get if AdBlock CDN processing is paused
// Returns: true if paused, false otherwise.
const isAdblockCDNPaused = function () {
  return LocalCDN.adblockCDNIsPaused();
};

// INFO ABOUT CURRENT PAGE

// Get interesting information about the current tab.
// Inputs:
//   callback: function(info).
//   info object passed to callback: {
//     tab: Tab object
//     whitelisted: bool - whether the current tab's URL is whitelisted.
//     disabled_site: bool - true if the url is e.g. about:blank or the
//                           Extension Gallery, where extensions don't run.
//     total_blocked: int - # of ads blocked since install
//     tab_blocked: int - # of ads blocked on this tab
//     display_stats: bool - whether block counts are displayed on button
//     display_menu_stats: bool - whether block counts are displayed on the popup menu
//   }
// Returns: null (asynchronous)
const getCurrentTabInfo = function (currentTabId, callback) {
  let localCurrentTabId = currentTabId;
  const processTab = function (tab) {
    if (!tab) {
      return;
    }
    const result = {
      tab,
    };

    if (!tab.url) {
      result.disabled_site = true;
      result.tab_blocked = 0;
    } else {
      result.tab.unicodeUrl = getUnicodeUrl(tab.url);
    }

    result.total_blocked = blockCounts.getTotalTrackersBlocked();
    result.total_HTTPS_redirects = blockCounts.getHTTPSRedirectTotal();
    result.total_redirect_count = LocalCDN.getRedirectCount();

    if (tab.url && tab.id) {
      result.tab_blocked = blockCounts.getTotalTrackersBlocked(tab.id);
      result.tab_HTTPS_redirects = blockCounts.getHTTPSRedirectTotal(tab.id);
      const currentTab = frameData.get(tab.id);
      if (currentTab) {
        result.tab_redirect_count = currentTab.redirectCount;
      } else {
        result.tab_redirect_count = 0;
      }
      if (typeof callback === 'function') {
        callback(result);
      }
    } else if (typeof callback === 'function') {
      callback(result);
    }
  };

  if (localCurrentTabId) {
    if (typeof localCurrentTabId === 'string') {
      localCurrentTabId = Number.parseInt(currentTabId, 10);
    }
    chrome.tabs.get(localCurrentTabId, processTab);
  } else {
    chrome.tabs.query({ active: true, currentWindow: true }, (tabs) => {
      if (!tabs) {
        return;
      }

      if (tabs.length === 0) {
        return;
      } // For example: only the background devtools or a popup are opened

      const tab = tabs[0];
      processTab(tab);
    });
  }
};

chrome.runtime.onMessage.addListener((request, sender, sendResponse) => {
  if (request.message !== 'getCurrentTabInfo') {
    return;
  }
  let currentTabId;
  if (request.tabId) {
    currentTabId = request.tabId;
  }
  getCurrentTabInfo(currentTabId, (info) => {
    if (info) {
      const updatedInfo = info;
      updatedInfo.settings = getSettings();
      updatedInfo.trackingPaused = adblockTrackingIsPaused();
      updatedInfo.cdnPaused = LocalCDN.adblockCDNIsPaused();
      updatedInfo.secureEnabled = DDGsettings.getSetting('httpsEverywhereEnabled');
      chrome.runtime.sendMessage({
        message: 'getCurrentTabInfoResponse',
        info: updatedInfo,
      });
    }
  });
  sendResponse({});
});

// BGcall DISPATCH
(function bgCallDispatch() {
  chrome.runtime.onMessage.addListener((request, sender, sendResponse) => {
    if (request.command !== 'call') {
      return;
    } // not for us
    const fn = window[request.fn];
    if (typeof fn !== 'function') {
      log('fn', fn, 'not found on background page, request: ', request);
      return;
    }
    request.args.push(sender);
    const result = fn.apply(window, request.args, sender);
    sendResponse(result);
  });
}());

// These functions are usually only called by content scripts.


// Attach methods to window
Object.assign(window, {
  adblockTrackingIsPaused,
  getSubscriptions,
  getSubscriptionsMinusText,
  getAdblockUserId,
  getDebugInfo,
  getFirstRun,
  handlerBehaviorChanged,
  openTab,
  setFirstRunToFalse,
  subscribe,
  unsubscribe,
  updateSubscriptionsNow,
});

initializeBackgroundObjects();

