'use strict';

/* For ESLint: List any global identifiers used in this file below */
/* global browser, log, Filter, determineUserLanguage, getUnicodeDomain,
FilterNormalizer, STATS, SelectorFilter, setDefault, handlerBehaviorChanged,
SnippetFilter,
BlockingFilterSet, FilterSet, getCustomFiltersText, idleHandler, getSettings */

// Requires jquery and must be on a page with access to the background page

// MyFilters class manages subscriptions and the FilterSet.

// Constructor: merge the stored subscription data and the total possible
// list of subscriptions into this.subscriptions.  Store to disk.
// Inputs: none.
const HOUR_IN_MS = 1000 * 60 * 60;
const keyPrefix = 'filter_list_';
const keyPrefixLength = keyPrefix.length;
const whiteSpaceRE = /\s/;

function MyFilters(callback) {
  const localThis = this;
  this.promise = new Promise(((resolve) => {
    localThis.readyComplete = resolve;
  }));

  this.getSubscriptions((subs) => {
    localThis.subscriptions = subs;

    localThis.officialOptions = localThis.makeSubscriptionOptions();
    if (typeof callback === 'function') {
      callback();
    }
  });
}

// Update _subscriptions and _official_options in case there are changes.
// Should be invoked right after creating a MyFilters object.
MyFilters.prototype.init = function myFiltersInit(callback) {
  this.updateDefaultSubscriptions();
  this.updateFieldsFromOriginalOptions();

  const localThis = this;
  // Build the filter list
  this.onSubscriptionChange(true);

  // On startup and then every hour, check if a list is out of date and has to
  // be updated
  if (this.newUser) {
    this.checkFilterUpdates();
  } else {
    idleHandler.scheduleItemOnce(
      () => {
        localThis.checkFilterUpdates();
      },

      60,
    );
  }

  window.setInterval(
    () => {
      idleHandler.scheduleItemOnce(() => {
        localThis.checkFilterUpdates();
      });
    },

    60 * 60 * 1000,
  );

  if (typeof callback === 'function') {
    callback();
  }
};

// Update the url and requiresList for entries in _subscriptions using values
// from _official_options.
MyFilters.prototype.updateFieldsFromOriginalOptions = function myFitlersUpdateFromOriginal() {
  // Use the stored properties, and only add any new properties and/or lists
  // if they didn't exist in this.subscriptions
  for (const id in this.officialOptions) {
    if (!this.subscriptions[id]) {
      this.subscriptions[id] = {};
    }
    const sub = this.subscriptions[id];
    const official = this.officialOptions[id];

    sub.initialUrl = sub.initialUrl || official.url;
    sub.url = sub.url || official.url;
    if (sub.initialUrl !== official.url) {
      // The official URL was changed. Use it. In case of a redirect, this
      // doesn't happen as only sub.url is changed, not sub.initialUrl.
      sub.initialUrl = official.url;
      sub.url = official.url;
    }

    const isMissingRequiredList = (sub.requiresList !== official.requiresList);
    if (official.requiresList && isMissingRequiredList && sub.subscribed) {
      // A required list was added.  Make sure main list subscribers get it.
      if (this.subscriptions[official.requiresList]) {
        this.changeSubscription(official.requiresList, { subscribed: true });
      }
    }

    sub.requiresList = official.requiresList;
    sub.subscribed = sub.subscribed || false;
    sub.allowSnippets = official.allowSnippets || false;
  }
};

// Update default subscriptions in the browser storage.
// Removes subscriptions localThis are no longer in the official list, not user
// submitted and no longer subscribed.
// Also, converts user submitted subscriptions to recognized one if it is already
// added to the official list and vice-versa.
MyFilters.prototype.updateDefaultSubscriptions = function myFiltersUpdateDefaultSubs() {
  if (this.newUser) {
    // Brand new user. Install some filters for them.
    this.subscriptions = this.loadDefaultSubscriptions();
    return;
  }

  for (const id in this.subscriptions) {
    // Delete unsubscribed ex-official lists.
    if (!this.officialOptions[id] && !this.subscriptions[id].userSubmitted
        && !this.subscriptions[id].subscribed) {
      delete this.subscriptions[id];
    } else {
      // Convert subscribed ex-official lists into user-submitted lists.
      // Convert subscribed ex-user-submitted lists into official lists.
      // Cache subscription that needs to be checked.
      const subToCheck = this.subscriptions[id];
      let isUserSubmitted = true;
      let updateID = id;
      if (!this.officialOptions[id]) {
        // If id is not in official options, check if there's a matching url in the
        // official list. If there is, then the subscription is not user submitted.
        for (const officialID in this.officialOptions) {
          const officialURL = this.officialOptions[officialID].url;
          if (subToCheck.initialUrl === officialURL
            || subToCheck.url === officialURL) {
            isUserSubmitted = false;
            updateID = officialID;
            break;
          }
        }
      } else {
        isUserSubmitted = false;
      }

      subToCheck.userSubmitted = isUserSubmitted;

      // Function that will add a new entry with updated id,
      // and will remove old entry with outdated id.
      const localThis = this;
      const renameSubscription = function (oldId, newId) {
        localThis.subscriptions[newId] = localThis.subscriptions[oldId];
        delete localThis.subscriptions[oldId];
      };

      // Create new id and check if new id is the same as id.
      // If not, update entry in subscriptions.
      const newID = isUserSubmitted ? (`url:${subToCheck.url}`) : updateID;

      if (newID !== id) {
        renameSubscription(id, newID);
      }
    }
  }
};

MyFilters.prototype.ready = function myFiltersReady() {
  return this.promise;
};
// When a subscription property changes, this function stores it
// Inputs: rebuild? boolean, true if the filterset should be rebuilt
MyFilters.prototype.onSubscriptionChange = function myFiltersSubChange(rebuild) {
  this.saveSubscriptions();

  // The only reasons to (re)build the filter set are
  // - when AdBlock starts
  // - when a filter list text is changed ([un]subscribed or updated a list)
  if (rebuild) {
    this.rebuild();
  }

  chrome.runtime.sendMessage({ command: 'filters_updated' });
};

// get filters that are defined in the extension
MyFilters.prototype.getExtensionFilters = function myFiltersExtensionFilters(settings) {
  const texts = [];
  if (settings.whitelist_hulu_ads) {
    // Issue 7178: FilterNormalizer removes EasyList's too-broad Hulu whitelist
    // entries.  If the user enables whitelist_hulu_ads, just add them back.
    // This workaround can be removed when EasyList changes its Hulu strategy.
    texts.push('@@||ads.hulu.com/published/*.flv');
    texts.push('@@||ads.hulu.com/published/*.mp4');
    texts.push('@@||ll.a.hulu.com/published/*.flv');
    texts.push('@@||ll.a.hulu.com/published/*.mp4');
  }

  return texts;
};

// Rebuild filters based on the current settings and subscriptions.
MyFilters.prototype.rebuild = function myFiltersRebuild() {
  let texts = [];

  // Only add subscriptions in Chrome, Opera, and older version of Safari...
  for (const id in this.subscriptions) {
    if (this.subscriptions[id].subscribed) {
      texts.push(this.subscriptions[id].text);
    }
  }

  // Include custom filters.
  const localThis = this;

  texts = texts.concat(localThis.getExtensionFilters(getSettings()));
  texts = texts.join('\n').split('\n');

  const filters = localThis.splitByType(texts);

  localThis.blocking = new BlockingFilterSet(
    FilterSet.fromFilters(filters.pattern),
    FilterSet.fromFilters(filters.whitelist),
  );

  handlerBehaviorChanged(); // defined in background

  localThis.readyComplete();

  // After 90 seconds, delete the cache. That way the cache is available when
  // rebuilding multiple times in a row (when multiple lists have to update at
  // the same time), but we save memory during all other times.
  window.setTimeout(() => {
    Filter.cache = {};
  }, 90000);
};

MyFilters.prototype.splitByType = function myFiltersSplitType(texts) {
  // Remove duplicates and empties.
  const unique = {};
  for (let i = 0; i < texts.length; i += 1) {
    unique[texts[i]] = 1;
  }

  delete unique[''];
  const hidingUnmerged = [];
  const advanceHidingUnmerged = [];
  const snippetsUnmerged = [];

  const filters = {
    hiding: {},
    exclude: {},
    advanceHiding: {},
    pattern: {},
    snippets: {},
    whitelist: {},
  };
  let filter;
  for (const text in unique) {
    filter = Filter.fromText(text);
    if (Filter.isSelectorExcludeFilter(text)) {
      setDefault(filters.exclude, filter.selector, []).push(filter);
    } else if (Filter.isSelectorFilter(text)) {
      hidingUnmerged.push(filter);
    } else if (Filter.isAdvancedSelectorFilter(text)) {
      advanceHidingUnmerged.push(filter);
    } else if (Filter.isContentFilter(text)) {
      snippetsUnmerged.push(filter);
    } else if (Filter.isWhitelistFilter(text)) {
      filters.whitelist[filter.id] = filter;
    } else {
      filters.pattern[filter.id] = filter;
    }
  }

  for (let i = 0; i < hidingUnmerged.length; i += 1) {
    filter = hidingUnmerged[i];
    const hider = SelectorFilter.merge(filter, filters.exclude[filter.selector]);
    filters.hiding[hider.id] = hider;
  }
  for (let i = 0; i < advanceHidingUnmerged.length; i += 1) {
    filter = advanceHidingUnmerged[i];
    const hider = SelectorFilter.merge(filter, filters.exclude[filter.selector]);
    filters.advanceHiding[hider.id] = hider;
  }
  for (let i = 0; i < snippetsUnmerged.length; i += 1) {
    filter = snippetsUnmerged[i];
    const snippet = SnippetFilter.merge(filter, filters.exclude[filter.selector]);
    filters.snippets[snippet.id] = snippet;
  }
  return filters;
};

// Change a property of a subscription or check if it has to be updated
// Inputs: id: the id of the subscription to change
//         subData: object containing all data that should be changed
//         forceFetch: if the subscriptions have to be fetched again forced
MyFilters.prototype.changeSubscription = function myFiltersChangeSub(id, subData, forceFetch) {
  let subscribeRequiredListToo = false;
  let listDidntExistBefore = false;

  // Check if the list has to be updated
  function outOfDate(subscription) {
    if (forceFetch) {
      return true;
    }

    // After a failure, wait at least a day to refetch (overridden below if
    // it's a new filter list, having no .text)
    const failedAt = subscription.lastUpdateFailedAt || 0;
    if (Date.now() - failedAt < HOUR_IN_MS * 24) {
      return false;
    }

    // Don't let expiresAfterHours delay indefinitely (Issue 7443)
    const hardStop = subscription.expiresAfterHoursHard || 240;
    const smallerExpiry = Math.min(subscription.expiresAfterHours, hardStop);
    const millis = Date.now() - subscription.lastUpdate;
    const returnVal = (millis > HOUR_IN_MS * smallerExpiry);
    return returnVal;
  }

  let newId = id;

  // Working with an unknown list: create the list entry
  if (!this.subscriptions[id]) {
    newId = this.customToDefaultId(id);
    if (/^url:.*/.test(newId)) {
      listDidntExistBefore = true;
      this.subscriptions[newId] = {
        userSubmitted: true,
        initialUrl: newId.substr(4),
        url: newId.substr(4),
        title: subData.title,
      };
    }

    subscribeRequiredListToo = true;
  }

  // Subscribing to a well known list should also subscribe to a required list
  if (!this.subscriptions[newId].subscribed && subData.subscribed) {
    subscribeRequiredListToo = true;
  }

  // Apply all changes from subData
  for (const property in subData) {
    if (subData[property] !== undefined) {
      this.subscriptions[newId][property] = subData[property];
    }
  }

  // Check if the required list is a well known list, but only if it is changed
  if (subData.requiresList) {
    // eslint-disable-next-line max-len
    this.subscriptions[newId].requiresList = this.customToDefaultId(this.subscriptions[newId].requiresList);
  }

  if (forceFetch) {
    delete this.subscriptions[newId].lastModified;
  }

  if (this.subscriptions[newId].subscribed) {
    if ((!this.subscriptions[newId].text)
         || outOfDate(this.subscriptions[newId])) {
      this.fetchAndUpdate(newId, listDidntExistBefore);
    }
  } else {
    // If unsubscribed, remove some properties
    delete this.subscriptions[newId].text;
    delete this.subscriptions[newId].lastUpdate;
    delete this.subscriptions[newId].expiresAfterHours;
    delete this.subscriptions[newId].lastUpdateFailedAt;
    delete this.subscriptions[newId].lastModified;
    if (this.subscriptions[newId].deleteMe) {
      // Need to remove the filter list from memory, and permanent storage
      delete this.subscriptions[newId];
      chrome.storage.local.remove(keyPrefix + newId);
    }
  }

  // Notify of change.  If we subscribed, we rebuilt above; so we
  // only force a rebuild if we unsubscribed.
  this.onSubscriptionChange(subData.subscribed === false);

  // Subscribe to a required list if nessecary
  if (subscribeRequiredListToo && this.subscriptions[newId]
    && this.subscriptions[newId].requiresList) {
    let requiredList = this.subscriptions[newId].requiresList;
    // If the required list is easylist, but the user is subscribed to easylist_lite
    // change the required list to easylist lite
    if (requiredList === 'easylist'
        && this.subscriptions.easylist_lite.subscribed) {
      requiredList = 'easylist_lite';
    }
    this.changeSubscription(requiredList, { subscribed: true });
  }
};

// Fetch a filter list and parse it
// id:        the id of the list
// isNewList: true when the list is completely new and must succeed or
//            otherwise it'll be deleted.
MyFilters.prototype.fetchAndUpdate = function myFiltersFetchUpdate(id, isNewList) {
  const { url } = this.subscriptions[id];
  const localThis = this;
  function onError() {
    localThis.subscriptions[id].lastUpdateFailedAt = Date.now();
    localThis.onSubscriptionChange();
    if (isNewList) {
      // Delete the list. The user subscribed to an invalid list URL.
      // Delay it a bit, so subscribe.html can first finish it's async call to
      // see if the subscription succeeded or not. Otherwise it assumes it was
      // a well-known subscription and will report a succesfull fetch
      window.setTimeout(() => {
        localThis.changeSubscription(id, { subscribed: false, deleteMe: true });
      }, 500);
    }
  }

  const ajaxRequest = {
    jsonp: false,
    url,
    cache: false,
    headers: {
      Accept: 'text/plain',
      'X-Client-ID': `AdBlock/${STATS.version}`,
      'If-Modified-Since': this.subscriptions[id].lastModified || undefined,
    },
    success(text, status, xhr) {
      // In case the subscription disappeared while we were out
      if (!localThis.subscriptions[id]
          || !localThis.subscriptions[id].subscribed) {
        return;
      }

      // Sometimes text is "". Happens sometimes.  Weird, I know.
      // Every legit list starts with a comment.
      if (status === 'notmodified') {
        log('List not modified ', id, url);
        localThis.updateSubscriptionText(id, localThis.subscriptions[id].text);
        localThis.onSubscriptionChange(true);
      } else if (text
                && (((typeof text === 'string')
                    && text.length !== 0 && Filter.isComment(text.trim()))
                  || (typeof text === 'object'))) {
        log('Fetched ', id, url);
        localThis.updateSubscriptionText(id, text, xhr);
        localThis.onSubscriptionChange(true);
      } else {
        log(`Fetched, but invalid list ${url}`);
        onError();
      }
    },

    error(xhr, textStatus, errorThrown) {
      if (localThis.subscriptions[id]) {
        onError();
      }
      log(`Error fetching ${url}`);
      log(`textStatus ${textStatus}`);
      log(`errorThrown ${errorThrown}`);
    },
  };
  $.ajax(ajaxRequest);
};

// Record that subscription_id is subscribed, was updated now, and has
// the given text.  Requires that this.subscriptions[subscription_id] exists.
// The xhr variable can be used to search the response headers
MyFilters.prototype.updateSubscriptionText = function myFiltersUpdateSubText(id, text, xhr) {
  this.subscriptions[id].lastUpdate = Date.now();
  delete this.subscriptions[id].lastUpdateFailedAt;

  // In case the resource wasn't modified, there is no need to reparse this.
  // xhr isn't send in this case. Do reparse .text, in case we had some update
  // which modified the checks in filternormalizer.js.
  if (xhr) {
    // Store the last time a resource was modified on the server, so we won't re-
    // fetch if it wasn't modified. It is null if the server doesn't support this.
    this.subscriptions[id].lastModified = xhr.getResponseHeader('Last-Modified');

    // Record how many hours until we need to update the subscription text. This
    // can be specified in the file. Defaults to 120.
    this.subscriptions[id].expiresAfterHours = 120;
    const checkLines = text.split('\n', 15); // 15 lines should be enough
    const expiresRegex = /(?:expires:|expires after ) *(\d+) ?(h?)/i;
    const redirectRegex = /(?:redirect:|redirects to ) *(https?:\/\/\S+)/i;
    for (let i = 0; i < checkLines.length; i += 1) {
      if (Filter.isComment(checkLines[i])) {
        let match = checkLines[i].match(redirectRegex);
        if (match && match[1] !== this.subscriptions[id].url) {
          [, this.subscriptions[id].url] = match; // assuming the URL is always correct

          // Force an update.  Even if our refetch below fails we'll have to
          // fetch the new URL in the future until it succeeds.
          this.subscriptions[id].lastUpdate = 0;
        }

        match = checkLines[i].match(expiresRegex);
        if (match && parseInt(match[1], 10)) {
          const hours = parseInt(match[1], 10) * (match[2] === 'h' ? 1 : 24);
          this.subscriptions[id].expiresAfterHours = Math.min(hours, 21 * 24); // 3 week maximum
        }
      }
    }

    // Smear expiry (Issue 7443)
    this.subscriptions[id].expiresAfterHoursHard = this.subscriptions[id].expiresAfterHours * 2;
    const smear = (Math.random() * 0.4) + 0.8;
    this.subscriptions[id].expiresAfterHours *= smear;
  }

  this.subscriptions[id].text = FilterNormalizer.normalizeList(text,
    false, this.subscriptions[id].allowSnippets);

  // The url changed. Simply refetch...
  if (this.subscriptions[id].lastUpdate === 0) {
    this.changeSubscription(id, {}, true);
  }
};

// Checks if subscriptions have to be updated
// Inputs: force? (boolean), true if every filter has to be updated
MyFilters.prototype.checkFilterUpdates = function myFiltersCheckFilterUpdates(force) {
  const key = 'last_subscriptions_check';
  const now = Date.now();
  const forceLocal = force;
  const localThis = this;
  chrome.storage.local.get(key, (response) => {
    const delta = now - (response[key] || now);
    const deltaHours = delta / HOUR_IN_MS;
    const saveData = {};
    saveData[key] = now;
    chrome.storage.local.set(saveData);
    if (deltaHours > 24) {
      // Extend expiration of subscribed lists (Issue 7443)
      for (const id in localThis.subscriptions) {
        if (localThis.subscriptions[id].subscribed) {
          localThis.subscriptions[id].expiresAfterHours += deltaHours;
        }
      }

      localThis.onSubscriptionChange(); // Store the change
    }

    for (const id in localThis.subscriptions) {
      if (localThis.subscriptions[id].subscribed) {
        localThis.changeSubscription(id, {}, forceLocal);
      }
    }
  });
};

// Checks if a custom id is of a known list
// Inputs: id: the list id to compare
// Returns the id that should be used
MyFilters.prototype.customToDefaultId = function myFiltersCustomToDefaultId(id) {
  const urlOfCustomList = id.substr(4);
  for (const defaultList in this.officialOptions) {
    if (this.officialOptions[defaultList].url === urlOfCustomList) {
      return defaultList;
    }
  }
  return id;
};

// If the user wasn't subscribed to any lists, subscribe to
// EasyPrivacy
// Inputs: none.
// Returns an object containing the subscribed lists
MyFilters.prototype.loadDefaultSubscriptions = function myFiltersLoadDefaultSubs() {
  const result = {};
  // Update will be done immediately after this function returns
  result.easyprivacy = { subscribed: true };
  return result;
};

// Used to create the list of default subscriptions
// Called when MyFilters is created.
// Returns: that list
MyFilters.prototype.makeSubscriptionOptions = function myFiltersMakeSubOptions() {
  // When modifying a list, IDs mustn't change!
  return {
    easyprivacy: { // EasyPrivacy
      url: 'https://easylist-downloads.adblockplus.org/easyprivacy.txt',
    }
  };
};

/* subscription properties:
url (string): url of subscription
initialUrl (string): the hardcoded url. Same as .url except when redirected
userSubmitted (bool): submitted by the user or not
requiresList (string): id of a list required for this list
subscribed (bool): if you are subscribed to the list or not
lastUpdate (date): time of the last succesfull update
lastModified (string): time of the last change on the server
lastUpdateFailedAt (date): if set, when the last update attempt failed
text (string): the filters of the subscription
expiresAfterHours (int): the time after which the subscription expires
expiresAfterHoursHard (int): we must redownload subscription after this delay
deleteMe (bool): if the subscription has to be deleted
allowSnippets (bool): if the subscription can have snippet rules
*/

// Save the current subscription information to storage
MyFilters.prototype.saveSubscriptions = function myFiltersSaveSubs() {
  const result = {};
  for (const id in this.subscriptions) {
    result[keyPrefix + id] = {};
    for (const attr in this.subscriptions[id]) {
      result[keyPrefix + id][attr] = this.subscriptions[id][attr];
    }
  }

  chrome.storage.local.set(result, () => {
    if (chrome.runtime.lastError) {
      log('write failure', chrome.runtime.lastError);
    }
  });
};

MyFilters.prototype.getSubscriptions = function myFiltersGetSubs(callback) {
  const localThis = this;
  chrome.storage.local.get(undefined, (readResponse) => {
    if (chrome.runtime.lastError) {
      log('read failure', chrome.runtime.lastError);
    }

    const readResult = {};
    for (const fullID in readResponse) {
      if (fullID.startsWith(keyPrefix)) {
        const id = fullID.substring(keyPrefixLength);
        readResult[id] = {};
        for (const attr in readResponse[fullID]) {
          readResult[id][attr] = readResponse[fullID][attr];
        }
      }
    }

    localThis.newUser = (Object.keys(readResult).length === 0);
    if (typeof callback === 'function') {
      callback(readResult);
    }
  });
};
